@file:JvmName("_IntConversions_Size")
@file:Suppress("NOTHING_TO_INLINE")

package ru.landrafhomyak.kotlin.utilities

import kotlin.contracts.InvocationKind
import kotlin.contracts.contract
import kotlin.jvm.JvmName


/********************************** extend *****************************************/

/**
 * Converts this [Byte] value to [Short].
 *
 * The resulting [Short] value represents the same numerical value as this [Byte].
 */
@JvmName("extendToShort")
public inline fun Byte.extendToShort(): Short = this.toShort()

/**
 * Converts this [Byte] value to [Int].
 *
 * The resulting [Int] value represents the same numerical value as this [Byte].
 */
@JvmName("extendToInt")
public inline fun Byte.extendToInt(): Int = this.toInt()

/**
 * Converts this [Byte] value to [Long].
 *
 * The resulting [Long] value represents the same numerical value as this [Byte].
 */
@JvmName("extendToLong")
public inline fun Byte.extendToLong(): Long = this.toLong()

/**
 * Converts this [UByte] value to [UShort].
 *
 * The resulting [UShort] value represents the same numerical value as this [UByte].
 */
@JvmName("extendToUShort")
public inline fun UByte.extendToUShort(): UShort = this.toUShort()

/**
 * Converts this [UByte] value to [UInt].
 *
 * The resulting [UInt] value represents the same numerical value as this [UByte].
 */
@JvmName("extendToUInt")
public inline fun UByte.extendToUInt(): UInt = this.toUInt()

/**
 * Converts this [UByte] value to [ULong].
 *
 * The resulting [ULong] value represents the same numerical value as this [UByte].
 */
@JvmName("extendToULong")
public inline fun UByte.extendToULong(): ULong = this.toULong()

/**
 * Converts this [Short] value to [Int].
 *
 * The resulting [Int] value represents the same numerical value as this [Short].
 */
@JvmName("extendToInt")
public inline fun Short.extendToInt(): Int = this.toInt()

/**
 * Converts this [Short] value to [Long].
 *
 * The resulting [Long] value represents the same numerical value as this [Short].
 */
@JvmName("extendToLong")
public inline fun Short.extendToLong(): Long = this.toLong()

/**
 * Converts this [UShort] value to [UInt].
 *
 * The resulting [UInt] value represents the same numerical value as this [UShort].
 */
@JvmName("extendToUInt")
public inline fun UShort.extendToUInt(): UInt = this.toUInt()

/**
 * Converts this [UShort] value to [ULong].
 *
 * The resulting [ULong] value represents the same numerical value as this [UShort].
 */
@JvmName("extendToULong")
public inline fun UShort.extendToULong(): ULong = this.toULong()

/**
 * Converts this [Int] value to [Long].
 *
 * The resulting [Long] value represents the same numerical value as this [Int].
 */
@JvmName("extendToLong")
public inline fun Int.extendToLong(): Long = this.toLong()

/**
 * Converts this [UInt] value to [ULong].
 *
 * The resulting [ULong] value represents the same numerical value as this [UInt].
 */
@JvmName("extendToULong")
public inline fun UInt.extendToULong(): ULong = this.toULong()

/******************** force trims *******************************/


/**
 * Converts this [Short] value to [Byte].
 *
 * If this value is in [Byte.MIN_VALUE]..[Byte.MAX_VALUE], the resulting [Byte] value represents
 * the same numerical value as this [Short].
 *
 * The resulting [Byte] value is represented by the least significant 8 bits of this [Short] value.
 */
@JvmName("trimToByte")
public inline fun Short.trimToByte(): Byte = this.toByte()


/**
 * Converts this [UShort] value to [UByte].
 *
 * If this value is in 0..[UByte.MAX_VALUE], the resulting [UByte] value represents
 * the same numerical value as this [UShort].
 *
 * The resulting [UByte] value is represented by the least significant 8 bits of this [UShort] value.
 */
@JvmName("trimToUByte")
public inline fun UShort.trimToUByte(): UByte = this.toUByte()


/**
 * Converts this [Int] value to [Byte].
 *
 * If this value is in [Byte.MIN_VALUE]..[Byte.MAX_VALUE], the resulting [Byte] value represents
 * the same numerical value as this [Int].
 *
 * The resulting [Byte] value is represented by the least significant 8 bits of this [Int] value.
 */
@JvmName("trimToByte")
public inline fun Int.trimToByte(): Byte = this.toByte()

/**
 * Converts this [Int] value to [Short].
 *
 * If this value is in [Short.MIN_VALUE]..[Short.MAX_VALUE], the resulting [Short] value represents
 * the same numerical value as this [Int].
 *
 * The resulting [Short] value is represented by the least significant 16 bits of this [Int] value.
 */
@JvmName("trimToShort")
public inline fun Int.trimToShort(): Short = this.toShort()


/**
 * Converts this [UInt] value to [UByte].
 *
 * If this value is in 0..[UByte.MAX_VALUE], the resulting [UByte] value represents
 * the same numerical value as this [UInt].
 *
 * The resulting [UByte] value is represented by the least significant 8 bits of this [UInt] value.
 */
@JvmName("trimToUByte")
public inline fun UInt.trimToUByte(): UByte = this.toUByte()

/**
 * Converts this [Int] value to [Short].
 *
 * If this value is in 0..[UShort.MAX_VALUE], the resulting [Short] value represents
 * the same numerical value as this [Int].
 *
 * The resulting [UShort] value is represented by the least significant 16 bits of this [UInt] value.
 */
@JvmName("trimToUShort")
public inline fun UInt.trimToUShort(): UShort = this.toUShort()


/**
 * Converts this [Long] value to [Byte].
 *
 * If this value is in [Byte.MIN_VALUE]..[Byte.MAX_VALUE], the resulting [Byte] value represents
 * the same numerical value as this [Long].
 *
 * The resulting [Byte] value is represented by the least significant 8 bits of this [Long] value.
 */
@JvmName("trimToByte")
public inline fun Long.trimToByte(): Byte = this.toByte()

/**
 * Converts this [Long] value to [Short].
 *
 * If this value is in [Short.MIN_VALUE]..[Short.MAX_VALUE], the resulting [Short] value represents
 * the same numerical value as this [Long].
 *
 * The resulting [Short] value is represented by the least significant 16 bits of this [Long] value.
 */
@JvmName("trimToShort")
public inline fun Long.trimToShort(): Short = this.toShort()

/**
 * Converts this [Long] value to [Int].
 *
 * If this value is in [Int.MIN_VALUE]..[Int.MAX_VALUE], the resulting [Int] value represents
 * the same numerical value as this [Long].
 *
 * The resulting [Int] value is represented by the least significant 32 bits of this [Long] value.
 */
@JvmName("trimToInt")
public inline fun Long.trimToInt(): Int = this.toInt()


/**
 * Converts this [ULong] value to [UByte].
 *
 * If this value is in 0..[UByte.MAX_VALUE], the resulting [UByte] value represents
 * the same numerical value as this [ULong].
 *
 * The resulting [UByte] value is represented by the least significant 8 bits of this [ULong] value.
 */
@JvmName("trimToUByte")
public inline fun ULong.trimToUByte(): UByte =this.toUByte()

/**
 * Converts this [ULong] value to [UShort].
 *
 * If this value is in 0..[UShort.MAX_VALUE], the resulting [UShort] value represents
 * the same numerical value as this [ULong].
 *
 * The resulting [UShort] value is represented by the least significant 16 bits of this [ULong] value.
 */
@JvmName("trimToUShort")
public inline fun ULong.trimToUShort(): UShort = this.toUShort()


/**
 * Converts this [ULong] value to [UInt].
 *
 * If this value is in 0..[UInt.MAX_VALUE], the resulting [UInt] value represents
 * the same numerical value as this [ULong].
 *
 * The resulting [UInt] value is represented by the least significant 32 bits of this [ULong] value.
 */
@JvmName("trimToUInt")
public inline fun ULong.trimToUInt(): UInt = this.toUInt()


/******************************** trims (or null) ****************************/


/**
 * Converts this [Short] value to [Byte].
 *
 * If this value is in [Byte.MIN_VALUE]..[Byte.MAX_VALUE], the resulting [Byte] value represents
 * the same numerical value as this [Short].
 * Otherwise `null` returned.
 */
@JvmName("toByteOrNull")
public inline fun Short.toByteOrNull(): Byte? = if (this in (Byte.MIN_VALUE.extendToShort())..(Byte.MAX_VALUE.extendToShort())) this.trimToByte() else null


/**
 * Converts this [UShort] value to [UByte].
 *
 * If this value is in 0..[UByte.MAX_VALUE], the resulting [UByte] value represents
 * the same numerical value as this [UShort].
 * Otherwise `null` returned.
 */
@JvmName("toUByteOrNull")
public inline fun UShort.toUByteOrNull(): UByte? = if (this <= UByte.MAX_VALUE.extendToUShort()) this.trimToUByte() else null


/**
 * Converts this [Int] value to [Byte].
 *
 * If this value is in [Byte.MIN_VALUE]..[Byte.MAX_VALUE], the resulting [Byte] value represents
 * the same numerical value as this [Int].
 * Otherwise `null` returned.
 */
@JvmName("toByteOrNull")
public inline fun Int.toByteOrNull(): Byte? = if (this in (Byte.MIN_VALUE.extendToInt())..(Byte.MAX_VALUE.extendToInt())) this.trimToByte() else null

/**
 * Converts this [Int] value to [Short].
 *
 * If this value is in [Short.MIN_VALUE]..[Short.MAX_VALUE], the resulting [Short] value represents
 * the same numerical value as this [Int].
 * Otherwise `null` returned.
 */
@JvmName("toShortOrNull")
public inline fun Int.toShortOrNull(): Short? = if (this in (Short.MIN_VALUE.extendToInt())..(Short.MAX_VALUE.extendToInt())) this.trimToShort() else null


/**
 * Converts this [UInt] value to [UByte].
 *
 * If this value is in 0..[UByte.MAX_VALUE], the resulting [UByte] value represents
 * the same numerical value as this [UInt].
 * Otherwise `null` returned.
 */
@JvmName("toUByteOrNull")
public inline fun UInt.toUByteOrNull(): UByte? = if (this <= UByte.MAX_VALUE.extendToUInt()) this.trimToUByte() else null

/**
 * Converts this [UInt] value to [Short].
 *
 * If this value is in 0..[UShort.MAX_VALUE], the resulting [UShort] value represents
 * the same numerical value as this [UInt].
 * Otherwise `null` returned.
 */
@JvmName("toUShortOrNull")
public inline fun UInt.toUShortOrNull(): UShort? = if (this <= UShort.MAX_VALUE.extendToUInt()) this.trimToUShort() else null


/**
 * Converts this [Long] value to [Byte].
 *
 * If this value is in [Byte.MIN_VALUE]..[Byte.MAX_VALUE], the resulting [Byte] value represents
 * the same numerical value as this [Long].
 * Otherwise `null` returned.
 */
@JvmName("toByteOrNull")
public inline fun Long.toByteOrNull(): Byte? = if (this in (Byte.MIN_VALUE.extendToLong())..(Byte.MAX_VALUE.extendToLong())) this.trimToByte() else null

/**
 * Converts this [Long] value to [Short].
 *
 * If this value is in [Short.MIN_VALUE]..[Short.MAX_VALUE], the resulting [Short] value represents
 * the same numerical value as this [Long].
 * Otherwise `null` returned.
 */
@JvmName("toShortOrNull")
public inline fun Long.toShortOrNull(): Short? = if (this in (Short.MIN_VALUE.extendToLong())..(Short.MAX_VALUE.extendToLong())) this.trimToShort() else null


/**
 * Converts this [Long] value to [Int].
 *
 * If this value is in [Int.MIN_VALUE]..[Int.MAX_VALUE], the resulting [Int] value represents
 * the same numerical value as this [Long].
 * Otherwise `null` returned.
 */
@JvmName("toIntOrNull")
public inline fun Long.toIntOrNull(): Int? = if (this in (Int.MIN_VALUE.extendToLong())..(Int.MAX_VALUE.extendToLong())) this.trimToInt() else null


/**
 * Converts this [ULong] value to [UByte].
 *
 * If this value is in 0..[UByte.MAX_VALUE], the resulting [UByte] value represents
 * the same numerical value as this [ULong].
 * Otherwise `null` returned.
 */
@JvmName("toUByteOrNull")
public inline fun ULong.toUByteOrNull(): UByte? = if (this in (UByte.MIN_VALUE.extendToULong())..(UByte.MAX_VALUE.extendToULong())) this.trimToUByte() else null

/**
 * Converts this [ULong] value to [UShort].
 *
 * If this value is in 0..[UShort.MAX_VALUE], the resulting [UShort] value represents
 * the same numerical value as this [ULong].
 * Otherwise `null` returned.
 */
@JvmName("toUShortOrNull")
public inline fun ULong.toUShortOrNull(): UShort? = if (this in (UShort.MIN_VALUE.extendToULong())..(UShort.MAX_VALUE.extendToULong())) this.trimToUShort() else null


/**
 * Converts this [ULong] value to [UInt].
 *
 * If this value is in 0..[UInt.MAX_VALUE], the resulting [UInt] value represents
 * the same numerical value as this [ULong].
 * Otherwise `null` returned.
 */
@JvmName("toUIntOrNull")
public inline fun ULong.toUIntOrNull(): UInt? = if (this in (UInt.MIN_VALUE.extendToULong())..(UInt.MAX_VALUE.extendToULong())) this.trimToUInt() else null


/********************************************* Conversion error formatters ****************************************/

@Suppress("FunctionName")
@JvmName("_formatTrimOverflowS")
private inline fun <reified W, reified N> _formatTrimOverflowS(value: W, min: N, max: N) =
	"Can't convert ${W::class.simpleName!!} value to ${N::class.simpleName!!} because $value !in ${min}..${max}"

@Suppress("FunctionName")
@JvmName("_formatTrimOverflowU")
private inline fun <reified W, reified N> _formatTrimOverflowU(value: W, max: N) =
	"Can't convert ${W::class.simpleName!!} value to ${N::class.simpleName!!} because $value > $max"


@JvmName("formatToByteError")
public fun Short.formatToByteError(): String = _formatTrimOverflowS<Short, Byte>(this, Byte.MIN_VALUE, Byte.MAX_VALUE)

@JvmName("formatToUByteError")
public fun UShort.formatToUByteError(): String = _formatTrimOverflowU<UShort, UByte>(this, UByte.MAX_VALUE)

@JvmName("formatToByteError")
public fun Int.formatToByteError(): String = _formatTrimOverflowS<Int, Byte>(this, Byte.MIN_VALUE, Byte.MAX_VALUE)

@JvmName("formatToShortError")
public fun Int.formatToShortError(): String = _formatTrimOverflowS<Int, Short>(this, Short.MIN_VALUE, Short.MAX_VALUE)

@JvmName("formatToUByteError")
public fun UInt.formatToUByteError(): String = _formatTrimOverflowU<UInt, UByte>(this, UByte.MAX_VALUE)

@JvmName("formatToUShortError")
public fun UInt.formatToUShortError(): String = _formatTrimOverflowU<UInt, UShort>(this, UShort.MAX_VALUE)

@JvmName("formatToByteError")
public fun Long.formatToByteError(): String = _formatTrimOverflowS<Long, Byte>(this, Byte.MIN_VALUE, Byte.MAX_VALUE)

@JvmName("formatToShortError")
public fun Long.formatToShortError(): String = _formatTrimOverflowS<Long, Short>(this, Short.MIN_VALUE, Short.MAX_VALUE)

@JvmName("formatToIntError")
public fun Long.formatToIntError(): String = _formatTrimOverflowS<Long, Int>(this, Int.MIN_VALUE, Int.MAX_VALUE)

@JvmName("formatToUByteError")
public fun ULong.formatToUByteError(): String = _formatTrimOverflowU<ULong, UByte>(this, UByte.MAX_VALUE)

@JvmName("formatToUShortError")
public fun ULong.formatToUShortError(): String = _formatTrimOverflowU<ULong, UShort>(this, UShort.MAX_VALUE)

@JvmName("formatToUIntError")
public fun ULong.formatToUIntError(): String = _formatTrimOverflowU<ULong, UInt>(this, UInt.MAX_VALUE)

/******************************** trims (or exception) ****************************/


/**
 * Converts this [Short] value to [Byte].
 *
 * If this value is in [Byte.MIN_VALUE]..[Byte.MAX_VALUE], the resulting [Byte] value represents
 * the same numerical value as this [Short].
 * Otherwise [exception] function called (by default throws [IllegalArgumentException]).
 */
@JvmName("toByteOrThrow")
public inline fun Short.toByteOrThrow(exception: (Short) -> Nothing = { b -> throw IllegalArgumentException(b.formatToByteError()) }): Byte {
	contract {
		callsInPlace(exception, InvocationKind.AT_MOST_ONCE)
	}

	return this.toByteOrNull() ?: exception(this)
}

/**
 * Converts this [UShort] value to [UByte].
 *
 * If this value is in 0..[UByte.MAX_VALUE], the resulting [UByte] value represents
 * the same numerical value as this [UShort].
 * Otherwise [exception] function called (by default throws [IllegalArgumentException]).
 */
@JvmName("toUByteOrThrow")
public inline fun UShort.toUByteOrThrow(exception: (UShort) -> Nothing = { b -> throw IllegalArgumentException(b.formatToUByteError()) }): UByte {
	contract {
		callsInPlace(exception, InvocationKind.AT_MOST_ONCE)
	}

	return this.toUByteOrNull() ?: exception(this)
}

/**
 * Converts this [Int] value to [Byte].
 *
 * If this value is in [Byte.MIN_VALUE]..[Byte.MAX_VALUE], the resulting [Byte] value represents
 * the same numerical value as this [Int].
 * Otherwise [exception] function called (by default throws [IllegalArgumentException]).
 */
@JvmName("toByteOrThrow")
public inline fun Int.toByteOrThrow(exception: (Int) -> Nothing = { b -> throw IllegalArgumentException(b.formatToByteError()) }): Byte {
	contract {
		callsInPlace(exception, InvocationKind.AT_MOST_ONCE)
	}

	return this.toByteOrNull() ?: exception(this)
}

/**
 * Converts this [Int] value to [Short].
 *
 * If this value is in [Short.MIN_VALUE]..[Short.MAX_VALUE], the resulting [Short] value represents
 * the same numerical value as this [Int].
 * Otherwise [exception] function called (by default throws [IllegalArgumentException]).
 */
@JvmName("toShortOrThrow")
public inline fun Int.toShortOrThrow(exception: (Int) -> Nothing = { b -> throw IllegalArgumentException(b.formatToShortError()) }): Short {
	contract {
		callsInPlace(exception, InvocationKind.AT_MOST_ONCE)
	}

	return this.toShortOrNull() ?: exception(this)
}

/**
 * Converts this [UInt] value to [UByte].
 *
 * If this value is in 0..[UByte.MAX_VALUE], the resulting [UByte] value represents
 * the same numerical value as this [UInt].
 * Otherwise [exception] function called (by default throws [IllegalArgumentException]).
 */
@JvmName("toUByteOrThrow")
public inline fun UInt.toUByteOrThrow(exception: (UInt) -> Nothing = { b -> throw IllegalArgumentException(b.formatToUByteError()) }): UByte {
	contract {
		callsInPlace(exception, InvocationKind.AT_MOST_ONCE)
	}

	return this.toUByteOrNull() ?: exception(this)
}


/**
 * Converts this [UInt] value to [UShort].
 *
 * If this value is in 0..[UShort.MAX_VALUE], the resulting [UShort] value represents
 * the same numerical value as this [UShort].
 * Otherwise [exception] function called (by default throws [IllegalArgumentException]).
 */
@JvmName("toUShortOrThrow")
public inline fun UInt.toUShortOrThrow(exception: (UInt) -> Nothing = { b -> throw IllegalArgumentException(b.formatToUShortError()) }): UShort {
	contract {
		callsInPlace(exception, InvocationKind.AT_MOST_ONCE)
	}

	return this.toUShortOrNull() ?: exception(this)
}


/**
 * Converts this [Long] value to [Byte].
 *
 * If this value is in [Byte.MIN_VALUE]..[Byte.MAX_VALUE], the resulting [Byte] value represents
 * the same numerical value as this [Long].
 * Otherwise [exception] function called (by default throws [IllegalArgumentException]).
 */
@JvmName("toByteOrThrow")
public inline fun Long.toByteOrThrow(exception: (Long) -> Nothing = { b -> throw IllegalArgumentException(b.formatToByteError()) }): Byte {
	contract {
		callsInPlace(exception, InvocationKind.AT_MOST_ONCE)
	}

	return this.toByteOrNull() ?: exception(this)
}

/**
 * Converts this [Long] value to [Short].
 *
 * If this value is in [Short.MIN_VALUE]..[Short.MAX_VALUE], the resulting [Short] value represents
 * the same numerical value as this [Long].
 * Otherwise [exception] function called (by default throws [IllegalArgumentException]).
 */
@JvmName("toShortOrThrow")
public inline fun Long.toShortOrThrow(exception: (Long) -> Nothing = { b -> throw IllegalArgumentException(b.formatToShortError()) }): Short {
	contract {
		callsInPlace(exception, InvocationKind.AT_MOST_ONCE)
	}

	return this.toShortOrNull() ?: exception(this)
}

/**
 * Converts this [Long] value to [Int].
 *
 * If this value is in [Int.MIN_VALUE]..[Int.MAX_VALUE], the resulting [Int] value represents
 * the same numerical value as this [Long].
 * Otherwise [exception] function called (by default throws [IllegalArgumentException]).
 */
@JvmName("toIntOrThrow")
public inline fun Long.toIntOrThrow(exception: (Long) -> Nothing = { b -> throw IllegalArgumentException(b.formatToIntError()) }): Int {
	contract {
		callsInPlace(exception, InvocationKind.AT_MOST_ONCE)
	}

	return this.toIntOrNull() ?: exception(this)
}

/**
 * Converts this [ULong] value to [UByte].
 *
 * If this value is in 0..[UByte.MAX_VALUE], the resulting [UByte] value represents
 * the same numerical value as this [ULong].
 * Otherwise [exception] function called (by default throws [IllegalArgumentException]).
 */
@JvmName("toUByteOrThrow")
public inline fun ULong.toUByteOrThrow(exception: (ULong) -> Nothing = { b -> throw IllegalArgumentException(b.formatToUByteError()) }): UByte {
	contract {
		callsInPlace(exception, InvocationKind.AT_MOST_ONCE)
	}

	return this.toUByteOrNull() ?: exception(this)
}


/**
 * Converts this [ULong] value to [UShort].
 *
 * If this value is in 0..[UShort.MAX_VALUE], the resulting [UShort] value represents
 * the same numerical value as this [ULong].
 * Otherwise [exception] function called (by default throws [IllegalArgumentException]).
 */
@JvmName("toUShortOrThrow")
public inline fun ULong.toUShortOrThrow(exception: (ULong) -> Nothing = { b -> throw IllegalArgumentException(b.formatToUShortError()) }): UShort {
	contract {
		callsInPlace(exception, InvocationKind.AT_MOST_ONCE)
	}

	return this.toUShortOrNull() ?: exception(this)
}


/**
 * Converts this [ULong] value to [UInt].
 *
 * If this value is in 0..[UInt.MAX_VALUE], the resulting [UInt] value represents
 * the same numerical value as this [ULong].
 * Otherwise [exception] function called (by default throws [IllegalArgumentException]).
 */
@JvmName("toUIntOrThrow")
public inline fun ULong.toUIntOrThrow(exception: (ULong) -> Nothing = { b -> throw IllegalArgumentException(b.formatToUIntError()) }): UInt {
	contract {
		callsInPlace(exception, InvocationKind.AT_MOST_ONCE)
	}

	return this.toUIntOrNull() ?: exception(this)
}

